export interface G3DHeader {
    magic: number; // MagicNumber GO3D[0x47 0x4f 0x33 0x44], GH3D [0x47 0x48 0x33 0x44] or CO3D[0x43 0x4f 0x33 0x44]
    width: number; // surface area silce width in pixels
    height: number; // surface area slice height in pixels
    depth: number; // surface area slice depth in pixels
    org_width: number; // silce width in pixels
    org_height: number; // slice height in pixels
    org_depth: number; // slice depth in pixels
    datatype: number; // 1 (byte per voxel) for gray, 3 (byte for voxel) for color
    unit: number; // km:3, m:0, mm:-3, micro m:-6, nano m:-9
    grid: number; // grid width in unit
    vs_width: number; // surface voxel width in unit
    v_width: number; // voxel width in unit
    v_height: number; // voxel height in unit
    v_depth: number; // voxel depth in unit
    ax: number; // rotation angle x
    ay: number; // rotation angle y
    az: number; // rotation angle z
    ox: number; // origin x in surface voxel coordinates
    oy: number; // origin y in surface voxel coordinates
    oz: number; // origin z in surface voxel coordinates
    s_num: number; // number of surface voxels
    s_th: number; // threshold of suface voxel
    option: string; // option area for additional parameters
    date: string; // free format - sample: 1970/01/01 00:00:00
    coordinate: string; // Talairach, Horsley-Clarke, etc
    subject: string; // subject name bm85... etc
    memo: string; // memo or something for extend
}

interface G3DImageRenderingContext {
    a1: number;
    a2: number;
    a3: number;
    b1: number;
    b2: number;
    b3: number;
    c1: number;
    c2: number;
    c3: number;
    d1: number;
    d2: number;
    d3: number;
    ra1: number;
    ra1_2: number;
    ra2: number;
    ra2_2: number;
    ra3: number;
    ra3_2: number;
    rb1: number;
    rb1_2: number;
    rb2: number;
    rb2_2: number;
    rb3: number;
    rb3_2: number;
    rc1: number;
    rc1_2: number;
    rc2: number;
    rc2_2: number;
    rc3: number;
    rc3_2: number;
    rd1: number;
    rd2: number;
    rd3: number;
}

interface G3DColor {
    r: number;
    g: number;
    b: number;
}

enum DrawingLineType {
    VLine = 0,
    VDottedLine = 1,
    HLine = 2,
    HDottedLine = 3,
}
const LineBrightness = 200;
const DottedLineBrightness = 128;
const EXACTSWITCH: 0 | 1 = 1;
const MaxIntensity = 200;

class G3D {
    private data: ArrayBuffer | null = null;
    private header: G3DHeader | null = null;
    private isLittleEndianFile = true;
    private surfaceView: Int16Array | null = null;
    private imgView: Uint8Array | null = null;
    private imageSize = 0;
    private centerX = 0;
    private centerY = 0;
    private centerZ = 0;
    private ox = 0;
    private oy = 0;
    private oz = 0;

    constructor(data: ArrayBuffer) {
        this._initialize(data);
    }

    getHeader(): G3DHeader | null {
        return this.header;
    }

    drawImage(canvas: HTMLCanvasElement, ax: number, ay: number, az: number, cutdepth: number, th: number, is3d: boolean, mesh: boolean, skincolor: boolean, smoothing: boolean): boolean {
        const irc: G3DImageRenderingContext = {
            a1: 0,
            a2: 0,
            a3: 0,
            b1: 0,
            b2: 0,
            b3: 0,
            c1: 0,
            c2: 0,
            c3: 0,
            d1: 0,
            d2: 0,
            d3: 0,
            ra1: 0,
            ra1_2: 0,
            ra2: 0,
            ra2_2: 0,
            ra3: 0,
            ra3_2: 0,
            rb1: 0,
            rb1_2: 0,
            rb2: 0,
            rb2_2: 0,
            rb3: 0,
            rb3_2: 0,
            rc1: 0,
            rc1_2: 0,
            rc2: 0,
            rc2_2: 0,
            rc3: 0,
            rc3_2: 0,
            rd1: 0,
            rd2: 0,
            rd3: 0,
        };
        // console.log(ax, ay, az, cutdepth, th, is3d, mesh, skincolor, smoothing);
        if (this.header === null) {
            return false;
        }

        canvas.width = this.imageSize;
        canvas.height = this.imageSize;
        const ctx = canvas.getContext('2d');
        if (ctx === null) {
            return false;
        }
        ctx.fillStyle = 'rgb(0, 0, 0)';
        ctx.fillRect(0, 0, this.imageSize, this.imageSize);
        const img = ctx.getImageData(0, 0, this.imageSize, this.imageSize);

        this._setMatrix(ax, ay, az, irc);
        switch (this.header.datatype) {
            case 1:
                if (is3d) {
                    const flag2d = [...Array<boolean>(this.imageSize * this.imageSize)].map(() => false);
                    const ddimg = [...Array<number>(this.imageSize * this.imageSize)].map(() => 0);
                    this._makeDIMG8(img, cutdepth, th, flag2d, ddimg, irc);
                    this._makeShadingImage(img, smoothing, false, flag2d, ddimg);
                } else {
                    this._makeSliceImage8(img, cutdepth, irc);
                }
                break;
            case 3:
                if (is3d) {
                    const flag2d = [...Array<boolean>(this.imageSize * this.imageSize)].map(() => false);
                    const ddimg = [...Array<number>(this.imageSize * this.imageSize)].map(() => 0);
                    this._makeDIMG24(img, cutdepth, th, flag2d, ddimg, irc);
                    this._makeShadingImage(img, smoothing, skincolor, flag2d, ddimg);
                } else {
                    this._makeSliceImage24(img, cutdepth, irc);
                }
                break;
        }
        if (mesh) {
            this._drawMesh(img, irc);
        }
        ctx.putImageData(img, 0, 0);
        return true;
    }

    printHeader(): void {
        if (this.header === null) {
            console.log('--- invalid g3d data ---');
            return;
        }
        let message = '--- g3d information ---\n';
        message += 'magic:              : ' + this.header.magic.toString(16) + '\n';
        message += 'width        [pixel]: ' + this.header.width.toString() + '\n';
        message += 'height       [pixel]: ' + this.header.height.toString() + '\n';
        message += 'depth        [pixel]: ' + this.header.depth.toString() + '\n';
        message += 'org_width    [pixel]: ' + this.header.org_width.toString() + '\n';
        message += 'org_height   [pixel]: ' + this.header.org_height.toString() + '\n';
        message += 'org_depth    [pixel]: ' + this.header.org_depth.toString() + '\n';
        message += 'datatype     [byte] : ' + this.header.datatype.toString() + '\n';
        message += 'unit         [m|E]  : ' + this.header.unit.toString() + '\n';
        message += 'grid         [unit] : ' + this.header.grid.toString() + '\n';
        message += 'surface voxel[unit] : ' + this.header.vs_width.toString() + '\n';
        message += 'voxel width  [unit] : ' + this.header.v_width.toString() + '\n';
        message += 'voxel height [unit] : ' + this.header.v_height.toString() + '\n';
        message += 'voxel depth  [unit] : ' + this.header.v_depth.toString() + '\n';
        message += 'ax           [deg]  : ' + this.header.ax.toString() + '\n';
        message += 'ay           [deg]  : ' + this.header.ay.toString() + '\n';
        message += 'az           [deg]  : ' + this.header.az.toString() + '\n';
        message += 'ox           [pixel]: ' + this.header.ox.toString() + '\n';
        message += 'oy           [pixel]: ' + this.header.oy.toString() + '\n';
        message += 'oz           [pixel]: ' + this.header.oz.toString() + '\n';
        message += 'ox_real      [unit] : ' + (this.header.ox * this.header.vs_width).toString() + '\n';
        message += 'oy_real      [unit] : ' + (this.header.oy * this.header.vs_width).toString() + '\n';
        message += 'oz_real      [unit] : ' + (this.header.oz * this.header.vs_width).toString() + '\n';
        message += 'surface voxels      : ' + this.header.s_num.toString() + '\n';
        message += 'threshold           : ' + this.header.s_th.toString() + '\n';
        message += 'date                : ' + this.header.date + '\n';
        message += 'coordinate          : ' + this.header.coordinate + '\n';
        message += 'subject             : ' + this.header.subject + '\n';
        message += 'memo                : ' + this.header.memo + '\n';
        message += 'File Type           : ' + (this.isLittleEndianFile ? 'Little Endian.' : 'Big Endian');
        console.log(message);
    }

    private _makeDIMG8(img: ImageData, cutdepth: number, th: number, flag2d: boolean[], ddimg: number[], irc: G3DImageRenderingContext) {
        if (this.header === null) {
            return;
        }
        const pdimg = [...Array<number>(this.imageSize * this.imageSize)].map(() => 0);
        const sw = th > this.header.s_th ? 2 : th < this.header.s_th ? 3 : EXACTSWITCH;
        switch (sw) {
            case 0: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        ddimg[pos] = pdimg[pos];
                        let pix = this._getPixel(x, y, cutdepth, irc);
                        if (pix > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        pix = this._getPixel(x, y, pdimg[pos], irc);
                        this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                    }
                }
                break;
            }
            case 1: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        let pdPos = pdimg[pos];
                        if (pdPos === 0) {
                            continue;
                        }
                        let pix = this._getPixel(x, y, cutdepth, irc);
                        if (pix > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        if (pdPos < cutdepth) {
                            pdPos = cutdepth;
                        }
                        ddimg[pos] = 0;
                        for (let z = pdPos; z < this.imageSize; z++) {
                            pix = this._getPixel(x, y, z, irc);
                            if (pix > th) {
                                ddimg[pos] = z;
                                this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                                break;
                            }
                        }
                    }
                }
                break;
            }
            case 2: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        let pdPos = pdimg[pos];
                        if (pdPos === 0) {
                            continue;
                        }
                        let pix = this._getPixel(x, y, cutdepth, irc);
                        if (pix > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        if (pix > this.header.s_th || pdPos < cutdepth) {
                            pdPos = cutdepth;
                        }
                        ddimg[pos] = 0;
                        for (let z = pdPos; z < this.imageSize; z++) {
                            pix = this._getPixel(x, y, z, irc);
                            if (pix > th) {
                                ddimg[pos] = z;
                                this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                                break;
                            }
                        }
                    }
                }
                break;
            }
            case 3: {
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        ddimg[pos] = 0;
                        for (let z = cutdepth; z < this.imageSize; z++) {
                            const pix = this._getPixel(x, y, z, irc);
                            if (pix > th) {
                                if (z === cutdepth) {
                                    flag2d[pos] = true;
                                } else {
                                    ddimg[pos] = z;
                                }
                                this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                                break;
                            }
                        }
                    }
                }
                break;
            }
        }
    }

    private _makeDIMG24(img: ImageData, cutdepth: number, th: number, flag2d: boolean[], ddimg: number[], irc: G3DImageRenderingContext) {
        if (this.header === null) {
            return;
        }
        const pdimg = [...Array<number>(this.imageSize * this.imageSize)].map(() => 0);
        const sw = th > this.header.s_th ? 2 : th < this.header.s_th ? 3 : EXACTSWITCH;
        switch (sw) {
            case 0: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        ddimg[pos] = pdimg[pos];
                        const c = this._getPixelColor(x, y, cutdepth, irc);
                        if (c.r > th || c.g > th || c.b > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, c);
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        this._drawPixel(img, pos, this._getPixelColor(x, y, pdimg[pos], irc));
                    }
                }
                break;
            }
            case 1: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let j = (pos = 0); j < this.imageSize; j++) {
                    for (let i = 0; i < this.imageSize; i++, pos++) {
                        let pdPos = pdimg[pos];
                        if (pdPos === 0) {
                            continue;
                        }
                        const c = this._getPixelColor(i, j, cutdepth, irc);
                        if (c.r > th || c.g > th || c.b > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, c);
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        if (pdPos < cutdepth) {
                            pdPos = cutdepth;
                        }
                        ddimg[pos] = 0;
                        for (let z = pdPos; z < this.imageSize; z++) {
                            const c = this._getPixelColor(i, j, z, irc);
                            if (c.r > th || c.g > th || c.b > th) {
                                ddimg[pos] = z;
                                this._drawPixel(img, pos, c);
                                break;
                            }
                        }
                    }
                }
                break;
            }
            case 2: {
                this._transSurface(cutdepth, pdimg, irc);
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        let pdPos = pdimg[pos];
                        if (pdPos === 0) {
                            continue;
                        }
                        const c = this._getPixelColor(x, y, cutdepth, irc);
                        if (c.r > th || c.g > th || c.b > th) {
                            flag2d[pos] = true;
                            this._drawPixel(img, pos, c);
                            ddimg[pos] = cutdepth;
                            continue;
                        }
                        if (c.r > this.header.s_th || c.g > this.header.s_th || c.b > this.header.s_th || pdPos < cutdepth) {
                            pdPos = cutdepth;
                        }
                        ddimg[pos] = 0;
                        for (let z = pdPos; z < this.imageSize; z++) {
                            const c = this._getPixelColor(x, y, z, irc);
                            if (c.r > th || c.g > th || c.b > th) {
                                ddimg[pos] = z;
                                this._drawPixel(img, pos, c);
                                break;
                            }
                        }
                    }
                }
                break;
            }
            case 3: {
                let pos = 0;
                for (let y = 0; y < this.imageSize; y++) {
                    for (let x = 0; x < this.imageSize; x++, pos++) {
                        ddimg[pos] = 0;
                        for (let z = cutdepth; z < this.imageSize; z++) {
                            const c = this._getPixelColor(x, y, z, irc);
                            if (c.r > th || c.g > th || c.b > th) {
                                if (z === cutdepth) {
                                    flag2d[pos] = true;
                                } else {
                                    ddimg[pos] = z;
                                }
                                this._drawPixel(img, pos, c);
                                break;
                            }
                        }
                    }
                }
                break;
            }
        }
    }

    private _makeShadingImage(img: ImageData, smoothing: boolean, skincolor: boolean, flag2d: boolean[], ddimg: number[]) {
        let pos = 0;
        for (let y = 0; y < this.imageSize; y++) {
            for (let x = 0; x < this.imageSize; x++, pos++) {
                let timg2d = 0;
                if (smoothing) {
                    if (x < this.imageSize - 2 && y < this.imageSize - 2) {
                        if (ddimg[pos] !== 0) {
                            const nx = (ddimg[pos] + ddimg[pos + this.imageSize] - ddimg[pos + 2] - ddimg[pos + 2 + this.imageSize]) / 4.0;
                            const ny = (ddimg[pos] + ddimg[pos + 1] - ddimg[pos + this.imageSize * 2] - ddimg[pos + 1 + this.imageSize * 2]) / 4.0;
                            timg2d = Math.round(MaxIntensity / Math.sqrt(nx * nx + ny * ny + 1));
                        }
                    }
                } else {
                    if (x < this.imageSize - 1 && y < this.imageSize - 1) {
                        if (ddimg[pos] !== 0) {
                            const nx = ddimg[pos] - ddimg[pos + 1];
                            const ny = ddimg[pos] - ddimg[pos + this.imageSize];
                            timg2d = Math.round(MaxIntensity / Math.sqrt(nx * nx + ny * ny + 1));
                        }
                    }
                }
                if (!flag2d[pos]) {
                    if (skincolor) {
                        const base = pos * 4;
                        const c: G3DColor = { r: img.data[base], g: img.data[base + 1], b: img.data[base + 2] };
                        const tmp = timg2d / Math.max(c.r, c.g, c.b);
                        this._drawPixel(img, pos, { r: c.r * tmp, g: c.g * tmp, b: c.b * tmp });
                    } else {
                        const pix = timg2d;
                        this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
                    }
                }
            }
        }
    }

    private _makeSliceImage8(img: ImageData, cutdepth: number, irc: G3DImageRenderingContext) {
        let pos = 0;
        for (let y = 0; y < this.imageSize; y++) {
            for (let x = 0; x < this.imageSize; x++, pos++) {
                const pix = this._getPixel(x, y, cutdepth, irc);
                this._drawPixel(img, pos, { r: pix, g: pix, b: pix });
            }
        }
    }

    private _makeSliceImage24(img: ImageData, cutdepth: number, irc: G3DImageRenderingContext) {
        let pos = 0;
        for (let y = 0; y < this.imageSize; y++) {
            for (let x = 0; x < this.imageSize; x++, pos++) {
                const c = this._getPixelColor(x, y, cutdepth, irc);
                this._drawPixel(img, pos, c);
            }
        }
    }

    private _drawMesh(img: ImageData, irc: G3DImageRenderingContext) {
        if (this.header === null) {
            return;
        }
        const origx = this._fround(this.header.ox * irc.ra1 + this.header.oy * irc.rb1 + this.header.oz * irc.rc1 + irc.rd1, 5);
        const origy = this._fround(this.header.ox * irc.ra2 + this.header.oy * irc.rb2 + this.header.oz * irc.rc2 + irc.rd2, 5);

        const band = this.header.grid / this.header.vs_width;
        const c = { r: 0, g: 0, b: LineBrightness };
        const cs = { r: 0, g: 0, b: DottedLineBrightness };
        this._drawLine(img, Math.round(origx), c, DrawingLineType.VLine);
        this._drawLine(img, Math.round(origy), c, DrawingLineType.HLine);
        for (let pos = Math.round(origx) + band; pos < this.imageSize; pos += band) {
            this._drawLine(img, Math.round(pos), cs, DrawingLineType.VDottedLine);
        }
        for (let pos = Math.round(origx) - band; pos >= 0; pos -= band) {
            this._drawLine(img, Math.round(pos), cs, DrawingLineType.VDottedLine);
        }
        for (let pos = Math.round(origy) + band; pos < this.imageSize; pos += band) {
            this._drawLine(img, Math.round(pos), cs, DrawingLineType.HDottedLine);
        }
        for (let pos = Math.round(origy) - band; pos >= 0; pos -= band) {
            this._drawLine(img, Math.round(pos), cs, DrawingLineType.HDottedLine);
        }
    }

    private _drawLine(img: ImageData, value: number, c: G3DColor, type: DrawingLineType) {
        switch (type) {
            case DrawingLineType.VLine: {
                for (let y = 0; y < this.imageSize; y++) {
                    this._drawPixel(img, value + y * this.imageSize, c);
                }
                break;
            }
            case DrawingLineType.VDottedLine: {
                for (let y = 0; y < this.imageSize; y += 2) {
                    this._drawPixel(img, value + y * this.imageSize, c);
                }
                break;
            }
            case DrawingLineType.HLine: {
                let pos = value * this.imageSize;
                for (let x = 0; x < this.imageSize; x++, pos++) {
                    this._drawPixel(img, pos, c);
                }
                break;
            }
            case DrawingLineType.HDottedLine: {
                let pos = value * this.imageSize;
                for (let x = 0; x < this.imageSize; x += 2, pos += 2) {
                    this._drawPixel(img, pos, c);
                }
                break;
            }
        }
    }

    private _drawPixel(img: ImageData, pos: number, c: G3DColor) {
        const base = pos * 4;
        img.data[base] = c.r;
        img.data[base + 1] = c.g;
        img.data[base + 2] = c.b;
        img.data[base + 3] = 255;
    }

    private _getPixelColor(x: number, y: number, z: number, irc: G3DImageRenderingContext): G3DColor {
        const c: G3DColor = { r: 0, g: 0, b: 0 };
        if (this.header === null || this.imgView === null) {
            return c;
        }

        const tx = x * irc.a1 + y * irc.b1 + z * irc.c1 + irc.d1;
        const ty = x * irc.a2 + y * irc.b2 + z * irc.c2 + irc.d2;
        const tz = x * irc.a3 + y * irc.b3 + z * irc.c3 + irc.d3;

        if (tx < 0 || ty < 0 || tz < 0) return c;
        if (tx > this.header.org_width - 1) return c;
        if (ty > this.header.org_height - 1) return c;
        if (tz > this.header.org_depth - 1) return c;

        const ix0 = Math.floor(tx);
        const iy0 = Math.floor(ty);
        const iz0 = Math.floor(tz);
        const wix1 = tx - ix0;
        const wix0 = 1.0 - wix1;
        const wiy1 = ty - iy0;
        const wiy0 = 1.0 - wiy1;
        const wiz1 = tz - iz0;
        const wiz0 = 1.0 - wiz1;
        const pixel = this.header.datatype;
        const line = this.header.org_width * pixel;
        const area = this.header.org_height * line;

        let offset = ix0 * pixel + iy0 * line + iz0 * area;
        let tmp1 = offset;
        let tmp2 = offset;
        const mx0 = [0, 0, 0];
        const mxy0 = [0, 0, 0];
        const mx1 = [0, 0, 0];
        const mxy1 = [0, 0, 0];

        if (ix0 + 1 === this.header.org_width) {
            tmp1 = offset;
            mx0[0] = this.imgView[tmp1++];
            mx0[1] = this.imgView[tmp1++];
            mx0[2] = this.imgView[tmp1];
        } else {
            tmp1 = offset;
            tmp2 = tmp1 + pixel;
            mx0[0] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            tmp1++;
            tmp2++;
            mx0[1] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            tmp1++;
            tmp2++;
            mx0[2] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
        }
        if (iy0 + 1 === this.header.org_height) {
            mxy0[0] = mx0[0];
            mxy0[1] = mx0[1];
            mxy0[2] = mx0[2];
        } else {
            if (ix0 + 1 === this.header.org_width) {
                tmp1 = offset + line;
                mx1[0] = this.imgView[tmp1++];
                mx1[1] = this.imgView[tmp1++];
                mx1[2] = this.imgView[tmp1];
            } else {
                tmp1 = offset + line;
                tmp2 = tmp1 + pixel;
                mx1[0] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
                tmp1++;
                tmp2++;
                mx1[1] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
                tmp1++;
                tmp2++;
                mx1[2] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            }
            mxy0[0] = mx0[0] * wiy0 + mx1[0] * wiy1;
            mxy0[1] = mx0[1] * wiy0 + mx1[1] * wiy1;
            mxy0[2] = mx0[2] * wiy0 + mx1[2] * wiy1;
        }
        if (iz0 + 1 === this.header.org_depth) {
            c.b = Math.round(mxy0[0]);
            c.g = Math.round(mxy0[1]);
            c.r = Math.round(mxy0[2]);
            return c;
        }
        offset = offset + area;
        if (ix0 + 1 === this.header.org_width) {
            tmp1 = offset;
            mx0[0] = this.imgView[tmp1++];
            mx0[1] = this.imgView[tmp1++];
            mx0[2] = this.imgView[tmp1];
        } else {
            tmp1 = offset;
            tmp2 = offset + pixel;
            mx0[0] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            tmp1++;
            tmp2++;
            mx0[1] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            tmp1++;
            tmp2++;
            mx0[2] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
        }
        if (iy0 + 1 === this.header.org_height) {
            mxy1[0] = mx0[0];
            mxy1[1] = mx0[1];
            mxy1[2] = mx0[2];
        } else {
            if (ix0 + 1 === this.header.org_width) {
                tmp1 = offset + line;
                mx1[0] = this.imgView[tmp1++];
                mx1[1] = this.imgView[tmp1++];
                mx1[2] = this.imgView[tmp1];
            } else {
                tmp1 = offset + line;
                tmp2 = tmp1 + pixel;
                mx1[0] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
                tmp1++;
                tmp2++;
                mx1[1] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
                tmp1++;
                tmp2++;
                mx1[2] = this.imgView[tmp1] * wix0 + this.imgView[tmp2] * wix1;
            }
            mxy1[0] = mx0[0] * wiy0 + mx1[0] * wiy1;
            mxy1[1] = mx0[1] * wiy0 + mx1[1] * wiy1;
            mxy1[2] = mx0[2] * wiy0 + mx1[2] * wiy1;
        }
        c.b = Math.round(mxy0[0] * wiz0 + mxy1[0] * wiz1);
        c.g = Math.round(mxy0[1] * wiz0 + mxy1[1] * wiz1);
        c.r = Math.round(mxy0[2] * wiz0 + mxy1[2] * wiz1);
        return c;
    }

    private _getPixel(x: number, y: number, z: number, irc: G3DImageRenderingContext): number {
        if (this.header === null || this.imgView === null) {
            return 0;
        }

        const tx = x * irc.a1 + y * irc.b1 + z * irc.c1 + irc.d1;
        const ty = x * irc.a2 + y * irc.b2 + z * irc.c2 + irc.d2;
        const tz = x * irc.a3 + y * irc.b3 + z * irc.c3 + irc.d3;

        if (tx < 0 || ty < 0 || tz < 0 || tx > this.header.org_width - 1 || ty > this.header.org_height - 1 || tz > this.header.org_depth - 1) {
            return 0;
        }

        const ix0 = Math.floor(tx);
        const iy0 = Math.floor(ty);
        const iz0 = Math.floor(tz);
        const wix1 = tx - ix0;
        const wix0 = 1.0 - wix1;
        const wiy1 = ty - iy0;
        const wiy0 = 1.0 - wiy1;
        const wiz1 = tz - iz0;
        const wiz0 = 1.0 - wiz1;
        const pixel = this.header.datatype;
        const line = this.header.org_width * pixel;
        const area = this.header.org_height * line;

        let mx0: number;
        let mx1: number;
        let mxy0: number;
        let mxy1: number;

        let offset = ix0 * pixel + iy0 * line + iz0 * area;
        if (ix0 + 1 === this.header.org_width) {
            mx0 = this.imgView[offset];
        } else {
            mx0 = this.imgView[offset] * wix0 + this.imgView[offset + pixel] * wix1;
        }
        if (iy0 + 1 === this.header.org_height) {
            mxy0 = mx0;
        } else {
            if (ix0 + 1 === this.header.org_width) {
                mx1 = this.imgView[offset + line];
            } else {
                mx1 = this.imgView[offset + line] * wix0 + this.imgView[offset + line + pixel] * wix1;
            }
            mxy0 = mx0 * wiy0 + mx1 * wiy1;
        }

        if (iz0 + 1 === this.header.org_depth) {
            return Math.round(mxy0);
        }

        offset = offset + area;
        if (ix0 + 1 === this.header.org_width) {
            mx0 = this.imgView[offset];
        } else {
            mx0 = this.imgView[offset] * wix0 + this.imgView[offset + pixel] * wix1;
        }
        if (iy0 + 1 === this.header.org_height) {
            mxy1 = mx0;
        } else {
            if (ix0 + 1 === this.header.org_width) {
                mx1 = this.imgView[offset + line];
            } else {
                mx1 = this.imgView[offset + line] * wix0 + this.imgView[offset + line + pixel] * wix1;
            }
            mxy1 = mx0 * wiy0 + mx1 * wiy1;
        }

        return Math.round(mxy0 * wiz0 + mxy1 * wiz1);
    }

    private _transSurface(cutdepth: number, pdimg: number[], irc: G3DImageRenderingContext) {
        if (this.header === null || this.surfaceView === null) {
            return;
        }

        const ix = [...Array<number>(7)].map(() => 0);
        const iy = [...Array<number>(7)].map(() => 0);
        const iz = [...Array<number>(7)].map(() => 0);

        for (let i = 0; i < this.header.s_num * 3; i++) {
            const x = this.surfaceView[i++];
            const y = this.surfaceView[i++];
            const z = this.surfaceView[i];

            ix[0] = x * irc.ra1 + y * irc.rb1 + z * irc.rc1 + irc.rd1;
            iy[0] = x * irc.ra2 + y * irc.rb2 + z * irc.rc2 + irc.rd2;
            iz[0] = x * irc.ra3 + y * irc.rb3 + z * irc.rc3 + irc.rd3;
            ix[1] = ix[0] + irc.ra1_2;
            iy[1] = iy[0] + irc.ra2_2;
            iz[1] = iz[0] + irc.ra3_2;
            ix[2] = ix[0] + irc.rb1_2;
            iy[2] = iy[0] + irc.rb2_2;
            iz[2] = iz[0] + irc.rb3_2;
            ix[3] = ix[0] + irc.rc1_2;
            iy[3] = iy[0] + irc.rc2_2;
            iz[3] = iz[0] + irc.rc3_2;
            ix[4] = ix[1] + irc.rb1_2;
            iy[4] = iy[1] + irc.rb2_2;
            iz[4] = iz[1] + irc.rb3_2;
            ix[5] = ix[1] + irc.rc1_2;
            iy[5] = iy[1] + irc.rc2_2;
            iz[5] = iz[1] + irc.rc3_2;
            ix[6] = ix[2] + irc.rc1_2;
            iy[6] = iy[2] + irc.rc2_2;
            iz[6] = iz[2] + irc.rc3_2;

            for (let j = 0; j < 7; j++) {
                const xi = Math.round(ix[j]);
                const yi = Math.round(iy[j]);
                if (xi < 0 || xi >= this.imageSize || yi < 0 || yi >= this.imageSize || iz[j] < cutdepth || iz[j] >= this.imageSize) {
                    continue;
                }
                const pos = xi + yi * this.imageSize;
                if (pdimg[pos] === 0 || pdimg[pos] > iz[j]) {
                    pdimg[pos] = iz[j];
                }
            }
        }
    }

    private _setMatrix(x: number, y: number, z: number, irc: G3DImageRenderingContext) {
        if (this.header === null) {
            return;
        }
        const xtr = [...Array<number>(9)].map(() => 0);
        const ytr = [...Array<number>(9)].map(() => 0);
        const ztr = [...Array<number>(9)].map(() => 0);
        const tmp = [...Array<number>(9)].map(() => 0);
        const tr = [...Array<number>(9)].map(() => 0);
        this._setXtr(xtr, (x * Math.PI) / 180.0);
        this._setYtr(ytr, (y * Math.PI) / 180.0);
        this._setZtr(ztr, (z * Math.PI) / 180.0);
        this._multi(ytr, xtr, tmp);
        this._multi(ztr, tmp, tr);
        irc.a1 = tr[0];
        irc.b1 = tr[1];
        irc.c1 = tr[2];
        irc.a2 = tr[3];
        irc.b2 = tr[4];
        irc.c2 = tr[5];
        irc.a3 = tr[6];
        irc.b3 = tr[7];
        irc.c3 = tr[8];
        irc.d1 = -irc.a1 * this.centerX - irc.b1 * this.centerY - irc.c1 * this.centerZ + this.ox;
        irc.d2 = -irc.a2 * this.centerX - irc.b2 * this.centerY - irc.c2 * this.centerZ + this.oy;
        irc.d3 = -irc.a3 * this.centerX - irc.b3 * this.centerY - irc.c3 * this.centerZ + this.oz;

        const sx = this.header.org_width / this.header.width;
        const sy = this.header.org_height / this.header.height;
        const sz = this.header.org_depth / this.header.depth;
        irc.a1 = sx * irc.a1;
        irc.b1 = sx * irc.b1;
        irc.c1 = sx * irc.c1;
        irc.d1 = sx * irc.d1;
        irc.a2 = sy * irc.a2;
        irc.b2 = sy * irc.b2;
        irc.c2 = sy * irc.c2;
        irc.d2 = sy * irc.d2;
        irc.a3 = sz * irc.a3;
        irc.b3 = sz * irc.b3;
        irc.c3 = sz * irc.c3;
        irc.d3 = sz * irc.d3;

        this._setXtr(xtr, (-x * Math.PI) / 180.0);
        this._setYtr(ytr, (-y * Math.PI) / 180.0);
        this._setZtr(ztr, (-z * Math.PI) / 180.0);
        this._multi(ytr, ztr, tmp);
        this._multi(xtr, tmp, tr);
        irc.ra1 = tr[0];
        irc.rb1 = tr[1];
        irc.rc1 = tr[2];
        irc.ra2 = tr[3];
        irc.rb2 = tr[4];
        irc.rc2 = tr[5];
        irc.ra3 = tr[6];
        irc.rb3 = tr[7];
        irc.rc3 = tr[8];
        irc.rd1 = -irc.ra1 * this.ox - irc.rb1 * this.oy - irc.rc1 * this.oz + this.centerX;
        irc.rd2 = -irc.ra2 * this.ox - irc.rb2 * this.oy - irc.rc2 * this.oz + this.centerY;
        irc.rd3 = -irc.ra3 * this.ox - irc.rb3 * this.oy - irc.rc3 * this.oz + this.centerZ;
        irc.ra1_2 = irc.ra1 / 2;
        irc.rb1_2 = irc.rb1 / 2;
        irc.rc1_2 = irc.rc1 / 2;
        irc.ra2_2 = irc.ra2 / 2;
        irc.rb2_2 = irc.rb2 / 2;
        irc.rc2_2 = irc.rc2 / 2;
        irc.ra3_2 = irc.ra3 / 2;
        irc.rb3_2 = irc.rb3 / 2;
        irc.rc3_2 = irc.rc3 / 2;
    }

    private _setXtr(xtr: number[], ang: number) {
        xtr[0] = 1;
        xtr[1] = 0;
        xtr[2] = 0;
        xtr[3] = 0;
        xtr[4] = Math.cos(ang);
        xtr[5] = -Math.sin(ang);
        xtr[6] = 0;
        xtr[7] = Math.sin(ang);
        xtr[8] = Math.cos(ang);
    }

    private _setYtr(ytr: number[], ang: number) {
        ytr[0] = Math.cos(ang);
        ytr[1] = 0;
        ytr[2] = Math.sin(ang);
        ytr[3] = 0;
        ytr[4] = 1;
        ytr[5] = 0;
        ytr[6] = -Math.sin(ang);
        ytr[7] = 0;
        ytr[8] = Math.cos(ang);
    }

    private _setZtr(ztr: number[], ang: number) {
        ztr[0] = Math.cos(ang);
        ztr[1] = -Math.sin(ang);
        ztr[2] = 0;
        ztr[3] = Math.sin(ang);
        ztr[4] = Math.cos(ang);
        ztr[5] = 0;
        ztr[6] = 0;
        ztr[7] = 0;
        ztr[8] = 1;
    }

    private _multi(l: number[], r: number[], a: number[]) {
        a[0] = l[0] * r[0] + l[1] * r[3] + l[2] * r[6];
        a[3] = l[3] * r[0] + l[4] * r[3] + l[5] * r[6];
        a[6] = l[6] * r[0] + l[7] * r[3] + l[8] * r[6];
        a[1] = l[0] * r[1] + l[1] * r[4] + l[2] * r[7];
        a[4] = l[3] * r[1] + l[4] * r[4] + l[5] * r[7];
        a[7] = l[6] * r[1] + l[7] * r[4] + l[8] * r[7];
        a[2] = l[0] * r[2] + l[1] * r[5] + l[2] * r[8];
        a[5] = l[3] * r[2] + l[4] * r[5] + l[5] * r[8];
        a[8] = l[6] * r[2] + l[7] * r[5] + l[8] * r[8];
    }

    private _fround(x: number, n: number) {
        return Math.round(x * Math.pow(10, n)) / Math.pow(10, n);
    }

    private _initialize(data: ArrayBuffer): boolean {
        // check browser endianness
        const isBigEndian = new Uint8Array(new Uint32Array([0x11223344]).buffer)[0] === 0x11;
        // check magic header and file endianness
        const view = new DataView(data);
        const magic = view.getInt32(0, true);
        let needSwap = false;
        switch (magic) {
            case 0x474f3344:
            case 0x47483344:
            case 0x434f3344:
                break;
            case 0x44334f47:
            case 0x44334847:
            case 0x44334f43:
                needSwap = true;
                break;
            default:
                console.log('invalid magic header found, data is not g3d format.');
                return false;
        }
        this.isLittleEndianFile = isBigEndian ? needSwap : !needSwap;
        // read header
        const decoder = new TextDecoder('utf-8');
        let headerOffset = 0;
        this.header = {
            magic: view.getInt32(headerOffset, this.isLittleEndianFile),
            width: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            height: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            depth: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            org_width: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            org_height: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            org_depth: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            datatype: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            unit: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            grid: view.getFloat64((headerOffset += 4), this.isLittleEndianFile),
            vs_width: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            v_width: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            v_height: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            v_depth: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            ax: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            ay: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            az: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            ox: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            oy: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            oz: view.getFloat64((headerOffset += 8), this.isLittleEndianFile),
            s_num: view.getInt32((headerOffset += 8), this.isLittleEndianFile),
            s_th: view.getInt32((headerOffset += 4), this.isLittleEndianFile),
            option: decoder.decode(new Uint8Array(data, (headerOffset += 4), 256)).replace(/\0+$/, ''),
            date: decoder.decode(new Uint8Array(data, (headerOffset += 256), 32)).replace(/\0+$/, ''),
            coordinate: decoder.decode(new Uint8Array(data, (headerOffset += 32), 32)).replace(/\0+$/, ''),
            subject: decoder.decode(new Uint8Array(data, (headerOffset += 32), 256)).replace(/\0+$/, ''),
            memo: decoder.decode(new Uint8Array(data, (headerOffset += 256), 256)).replace(/\0+$/, ''),
        };
        // set surface and image views
        const surfaceOffset = 964;
        const surfaceLength = this.header.s_num * 3;
        this.surfaceView = new Int16Array(data, surfaceOffset, surfaceLength);
        if (needSwap) {
            // fix endianness
            let offset = surfaceOffset;
            for (let i = 0; i < surfaceLength; i++) {
                const ui8 = new Uint8Array(data, offset, 2);
                ui8.reverse();
                offset += 2;
            }
        }
        const imgOffset = surfaceOffset + surfaceLength * 2;
        const imgLength = this.header.org_width * this.header.org_height * this.header.org_depth * this.header.datatype;
        this.imgView = new Uint8Array(data, imgOffset, imgLength);
        // set offset for image rendering
        this.imageSize = Math.ceil(Math.sqrt(this.header.width * this.header.width + this.header.height * this.header.height + this.header.depth * this.header.depth)) + 1;
        this.centerX = (this.imageSize - 1) / 2.0;
        this.centerY = (this.imageSize - 1) / 2.0;
        this.centerZ = (this.imageSize - 1) / 2.0;
        this.ox = (this.header.width - 1) / 2.0;
        this.oy = (this.header.height - 1) / 2.0;
        this.oz = (this.header.depth - 1) / 2.0;
        // set data
        this.data = data;
        return true;
    }
}

export default G3D;
